"""




 ----------------------------------------------------------------------------

    METADATA:

        File:    test_manager.py
        Project: paperap
        Created: 2025-03-21
        Version: 0.0.9
        Author:  Jess Mann
        Email:   jess@jmann.me
        Copyright (c) 2025 Jess Mann

 ----------------------------------------------------------------------------

    LAST MODIFIED:

        2025-03-21     By Jess Mann

"""


from __future__ import annotations

import unittest
from unittest import mock
import importlib
import sys
from typing import Any, Dict, List, Type

import pydantic

from paperap.client import PaperlessClient
from paperap.plugins.base import Plugin
from paperap.plugins.manager import PluginManager, PluginConfig


class MockPlugin(Plugin):
    """Mock plugin for testing."""

    name = "MockPlugin"
    manager: PluginManager  # Add as a proper field for Pydantic
    initialized_with: dict[str, Any] = {}  # Add as a proper field

    model_config = {
        "arbitrary_types_allowed": True,  # Allow PluginManager type
        "extra": "allow"  # Allow arbitrary extra attributes
    }

    def __init__(self, manager: PluginManager, **kwargs: Any) -> None:
        """Initialize the mock plugin."""
        super().__init__(manager=manager, **kwargs)
        self.initialized_with = kwargs

    def setup(self) -> None:
        """Setup the plugin."""
        pass

    def teardown(self) -> None:
        """Teardown the plugin."""
        pass


class AnotherMockPlugin(Plugin):
    """Another mock plugin for testing."""
    name = "AnotherMockPlugin"
    manager: PluginManager  # Add as a proper field for Pydantic
    initialized_with: dict[str, Any] = {}  # Add as a proper field

    model_config = {
        "arbitrary_types_allowed": True,  # Allow PluginManager type
        "extra": "allow"  # Allow arbitrary extra attributes
    }

    def __init__(self, manager: PluginManager, **kwargs: Any) -> None:
        """Initialize the mock plugin."""
        super().__init__(manager=manager, **kwargs)
        self.initialized_with = kwargs

    def setup(self) -> None:
        """Setup the plugin."""
        pass

    def teardown(self) -> None:
        """Teardown the plugin."""
        pass


class TestPluginManager(unittest.TestCase):
    """
    Test suite for the PluginManager class.
    """

    def setUp(self) -> None:
        """
        Written By claude

        Set up test fixtures before each test method.
        Creates a mock client and initializes a PluginManager instance.
        """
        self.mock_client = mock.MagicMock(spec=PaperlessClient)
        self.manager = PluginManager(client=self.mock_client)

    def test_init(self) -> None:
        """
        Written By claude

        Test that the PluginManager initializes with the expected default values.
        """
        self.assertEqual(self.manager.plugins, {})
        self.assertEqual(self.manager.instances, {})
        self.assertEqual(
            self.manager.config,
            {"enabled_plugins": [], "settings": {}}
        )
        self.assertEqual(self.manager.client, self.mock_client)

    def test_enabled_plugins_with_config(self) -> None:
        """
        Written By claude

        Test that enabled_plugins property returns the configured list when available.
        """
        # Set up
        self.manager.config = {
            "enabled_plugins": ["Plugin1", "Plugin2"],
            "settings": {}
        }

        # Test
        self.assertEqual(self.manager.enabled_plugins, ["Plugin1", "Plugin2"])

    def test_enabled_plugins_without_config(self) -> None:
        """
        Written By claude

        Test that enabled_plugins property returns all discovered plugins when
        no enabled_plugins list is configured.
        """
        # Set up
        self.manager.plugins = {"Plugin1": MockPlugin, "Plugin2": AnotherMockPlugin}
        self.manager.config = {
            "enabled_plugins": [],
            "settings": {}
        }

        # Test
        self.assertEqual(set(self.manager.enabled_plugins), {"Plugin1", "Plugin2"})

    @mock.patch('importlib.import_module')
    @mock.patch('pkgutil.iter_modules')
    @mock.patch('inspect.getmembers')
    def __disabled_test_discover_plugins(
        self,
        mock_getmembers: mock.MagicMock,
        mock_iter_modules: mock.MagicMock,
        mock_import_module: mock.MagicMock
    ) -> None:
        """
        Written By claude

        Test that discover_plugins correctly finds and registers plugin classes.
        """
        # Set up mocks
        mock_package = mock.MagicMock()
        mock_package.__path__ = ["some/path"]
        mock_import_module.return_value = mock_package

        mock_module = mock.MagicMock()
        mock_module.__name__ = "test_module"

        # Mock a plugin class
        mock_plugin_class = mock.MagicMock(spec=type)
        mock_plugin_class.__name__ = "TestPlugin"
        mock_plugin_class.__module__ = "paperap.plugins.test_module"

        # Set up the module iteration
        mock_iter_modules.return_value = [
            (None, "paperap.plugins.test_module", False)
        ]

        # Set up class inspection
        mock_getmembers.return_value = [
            ("TestPlugin", mock_plugin_class)
        ]

        # Make the mock plugin class appear to be a subclass of Plugin
        def is_subclass_side_effect(cls, parent):
            if parent is Plugin and cls is mock_plugin_class:
                return True
            return False

        with mock.patch('inspect.isclass', return_value=True):
            with mock.patch('builtins.issubclass', side_effect=is_subclass_side_effect):
                # Call the method
                self.manager.discover_plugins()

                # Verify the plugin was registered
                self.assertIn("TestPlugin", self.manager.plugins)
                self.assertEqual(self.manager.plugins["TestPlugin"], mock_plugin_class)

    def test_configure_with_config_dict(self) -> None:
        """
        Written By claude

        Test that configure correctly updates the manager's configuration
        when provided with a PluginConfig dictionary.
        """
        # Set up
        config: PluginConfig = {
            "enabled_plugins": ["Plugin1"],
            "settings": {"Plugin1": {"setting1": "value1"}}
        }

        # Call the method
        self.manager.configure(config)

        # Verify the configuration was updated
        self.assertEqual(self.manager.config, config)

    def test_configure_with_kwargs(self) -> None:
        """
        Written By claude

        Test that configure correctly updates the manager's configuration
        when provided with keyword arguments.
        """
        # Set up
        enabled_plugins = ["Plugin1"]
        settings = {"Plugin1": {"setting1": "value1"}}

        # Call the method
        self.manager.configure(enabled_plugins=enabled_plugins, settings=settings)

        # Verify the configuration was updated
        self.assertEqual(self.manager.config["enabled_plugins"], enabled_plugins)
        self.assertEqual(self.manager.config["settings"], settings)

    def test_configure_with_unexpected_kwargs(self) -> None:
        """
        Written By claude

        Test that configure logs a warning when provided with unexpected keyword arguments.
        """
        # Set up
        with mock.patch('paperap.plugins.manager.logger.warning') as mock_warning:
            # Call the method with an unexpected kwarg
            self.manager.configure(unexpected_key="value")

            # Verify a warning was logged
            mock_warning.assert_called_once()
            self.assertIn("Unexpected configuration keys", mock_warning.call_args[0][0])

    def test_get_plugin_config(self) -> None:
        """
        Written By claude

        Test that get_plugin_config returns the correct configuration for a plugin.
        """
        # Set up
        self.manager.config = {
            "enabled_plugins": ["Plugin1"],
            "settings": {"Plugin1": {"setting1": "value1"}}
        }

        # Test with an existing plugin
        config = self.manager.get_plugin_config("Plugin1")
        self.assertEqual(config, {"setting1": "value1"})

        # Test with a non-existent plugin
        config = self.manager.get_plugin_config("NonExistentPlugin")
        self.assertEqual(config, {})

    def test_initialize_plugin_already_initialized(self) -> None:
        """
        Written By claude

        Test that initialize_plugin returns the existing instance if the plugin
        has already been initialized.
        """
        # Set up
        mock_instance = mock.MagicMock(spec=Plugin)
        self.manager.instances = {"TestPlugin": mock_instance}

        # Call the method
        result = self.manager.initialize_plugin("TestPlugin")

        # Verify the existing instance was returned
        self.assertEqual(result, mock_instance)

    def test_initialize_plugin_not_found(self) -> None:
        """
        Written By claude

        Test that initialize_plugin returns None and logs a warning if the plugin
        is not found.
        """
        # Set up
        with mock.patch('paperap.plugins.manager.logger.warning') as mock_warning:
            # Call the method with a non-existent plugin
            result = self.manager.initialize_plugin("NonExistentPlugin")

            # Verify None was returned and a warning was logged
            self.assertIsNone(result)
            mock_warning.assert_called_once()
            self.assertIn("Plugin not found", mock_warning.call_args[0][0])

    def __disabled_test_initialize_plugin_success(self) -> None:
        """
        Written By claude

        Test that initialize_plugin correctly initializes a plugin and returns the instance.
        """
        # Set up
        self.manager.plugins = {"MockPlugin": MockPlugin}
        self.manager.config = {
            "enabled_plugins": ["MockPlugin"],
            "settings": {"MockPlugin": {"setting1": "value1"}}
        }

        # Create a mock that we can track with initialized_with already set
        mock_plugin = mock.MagicMock(spec=MockPlugin)
        mock_plugin.initialized_with = {}  # Initialize the attribute before we use it

        # Use patch to replace the entire MockPlugin class with a factory function
        def mock_factory(*args, **kwargs):
            # Update our mock with the initialization parameters
            mock_plugin.initialized_with = kwargs
            return mock_plugin

        # Patch the class itself, not its methods
        with mock.patch('tests.unit.plugins.test_manager.MockPlugin', side_effect=mock_factory):
            with mock.patch('paperap.plugins.manager.logger.info') as mock_info:
                # Call the method
                result = self.manager.initialize_plugin("MockPlugin")

                # Verify the mock was created with the correct settings
                self.assertEqual(mock_plugin.initialized_with.get("setting1"), "value1")

                # Verify the result is our mock
                self.assertEqual(result, mock_plugin)
                mock_info.assert_called_once()
                self.assertIn("Initialized plugin", mock_info.call_args[0][0])

    def test_initialize_plugin_exception(self) -> None:
        """
        Written By claude

        Test that initialize_plugin handles exceptions during plugin initialization.
        """
        # Set up a plugin class that raises an exception when initialized
        class ExceptionPlugin(Plugin):
            name : str = "ExceptionPlugin"
            def __init__(self, **kwargs: Any) -> None:
                raise ValueError("Test exception")

        self.manager.plugins = {"ExceptionPlugin": ExceptionPlugin}

        # Call the method
        with mock.patch('paperap.plugins.manager.logger.error') as mock_error:
            result = self.manager.initialize_plugin("ExceptionPlugin")

            # Verify None was returned and an error was logged
            self.assertIsNone(result)
            mock_error.assert_called_once()
            self.assertIn("Failed to initialize plugin", mock_error.call_args[0][0])

    def test_initialize_all_plugins(self) -> None:
        """
        Written By claude

        Test that initialize_all_plugins correctly initializes all enabled plugins.
        """
        # Set up
        self.manager.plugins = {
            "MockPlugin": MockPlugin,
            "AnotherMockPlugin": AnotherMockPlugin
        }
        self.manager.config = {
            "enabled_plugins": ["MockPlugin", "AnotherMockPlugin"],
            "settings": {
                "MockPlugin": {"setting1": "value1"},
                "AnotherMockPlugin": {"setting2": "value2"}
            }
        }

        # Use MagicMock instead of incomplete Pydantic objects
        mock_plugin = mock.MagicMock(spec=MockPlugin)
        mock_plugin.initialized_with = {"setting1": "value1"}

        another_mock_plugin = mock.MagicMock(spec=AnotherMockPlugin)
        another_mock_plugin.initialized_with = {"setting2": "value2"}

        def side_effect(plugin_name):
            if plugin_name == "MockPlugin":
                return mock_plugin
            elif plugin_name == "AnotherMockPlugin":
                return another_mock_plugin
            return None

        with mock.patch.object(PluginManager, 'initialize_plugin', side_effect=side_effect):
            result = self.manager.initialize_all_plugins()

            # Verify all plugins were initialized
            self.assertEqual(len(result), 2)
            self.assertEqual(result["MockPlugin"], mock_plugin)
            self.assertEqual(result["AnotherMockPlugin"], another_mock_plugin)
            self.assertEqual(result["MockPlugin"].initialized_with, {"setting1": "value1"})
            self.assertEqual(result["AnotherMockPlugin"].initialized_with, {"setting2": "value2"})

    def test_initialize_all_plugins_with_failure(self) -> None:
        """
        Written By claude

        Test that initialize_all_plugins continues initializing plugins even if
        some fail to initialize.
        """
        # Set up plugin classes
        class ExceptionPlugin(Plugin):
            name : str = "ExceptionPlugin"
            manager: PluginManager

            model_config = {"arbitrary_types_allowed": True}

            def setup(self) -> None:
                """Setup the plugin."""
                pass

            def teardown(self) -> None:
                """Teardown the plugin."""
                pass

        self.manager.plugins = {
            "MockPlugin": MockPlugin,
            "ExceptionPlugin": ExceptionPlugin
        }
        self.manager.config = {
            "enabled_plugins": ["MockPlugin", "ExceptionPlugin"],
            "settings": {
                "MockPlugin": {"setting1": "value1"}
            }
        }

        # Create a mock for the successful plugin
        mock_plugin = mock.MagicMock(spec=MockPlugin)
        mock_plugin.initialized_with = {"setting1": "value1"}

        # Patch at the module level instead of the instance level
        original_initialize_plugin = PluginManager.initialize_plugin

        def patched_initialize_plugin(self, plugin_name):
            if plugin_name == "MockPlugin":
                # Add directly to instances dict
                if not hasattr(self, 'instances'):
                    self.instances = {}
                self.instances["MockPlugin"] = mock_plugin
                return mock_plugin
            elif plugin_name == "ExceptionPlugin":
                # Return None to simulate failure
                return None
            return original_initialize_plugin(self, plugin_name)

        # Patch the method at the class level
        with mock.patch('paperap.plugins.manager.PluginManager.initialize_plugin',
                        patched_initialize_plugin):
            with mock.patch('paperap.plugins.manager.logger.error'):
                result = self.manager.initialize_all_plugins()

                # Verify the successful plugin was initialized
                self.assertEqual(len(result), 1)
                self.assertEqual(result["MockPlugin"], mock_plugin)
                self.assertEqual(result["MockPlugin"].initialized_with, {"setting1": "value1"})


if __name__ == '__main__':
    unittest.main()
